#!/usr/bin/python
'''
Assessment of the protein-RNA complexes and clustering the best solutions
'''
__author__ = 'Irina Tuszynska'
__email__ = 'irena@genesilico.pl'

import pickle, os, Numeric, optparse, sys
from Bio.PDB import Structure, Model, Chain, Residue, Atom, NeighborSearch, Vector
from genesilico.structure.BrutePDBParser import BrutePDBParser
from math import log, sqrt, ceil, degrees, pi, acos
from numpy import array, matrix, zeros

## read/validate command-line arguments
optparser=optparse.OptionParser(usage="%prog [<options>]")
optparser.add_option('-f',type="string",
                     dest="List_of_structures",
                     default='',
                     help="A file with list of PDB files. One path per line")
optparser.add_option('-s',type="string",
                     dest="Structure",
                     default='',
                     help="A filename with structure of protein-RNA complex")
optparser.add_option('-m',type="int",
                     dest="Nr_to_clust",
                     default=100,
                     help="Number best scored structures to clustering (default 100)")
optparser.add_option('-c',type="float",
                     dest="Cut_off",
                     default=5.0,
                     help="Threeshold in angstrems for clustering best scored decoys. Default value = 5")
optparser.add_option('-n',
                     dest="No_clustering",
		     action = "store_true",
		     default = False,
                     help="Only assess structures without clustering best scored results.")


(opts, args)=optparser.parse_args()

if len(sys.argv) == 1:
    print optparser.format_help() #prints help if no arguments
    sys.exit(1)

if opts.List_of_structures and opts.Structure:
    optparser.error("options -f and -s are mutually exclusive")

if opts.No_clustering:
    opts.Nr_to_clust = 0

if opts.Nr_to_clust and opts.No_clustering:
    optparser.error("options -m and -n are mutually exclusive")




class RedProt():
    #class RedProt(PDBParser):
  '''convert protein structure to reduced model according to MB. '''
  def __init__(self):
    self.S1 = {'ALA':['CB',1.0], 'SER':['CB','OG',2.0], 'CYS':['CB','SG',2.0], 'VAL':['CB','CG1','CG2',3.0], 'THR':['CB','OG1','CG2',3.0], 
		'ILE':['CB','CG1','CG2','CD1',4.0], 'PRO':['CB','CG','CD',3.0], 'MET':['CB','CG',2.0], 'ASP':['CB',1.0], 'ASN':['CB',1.0], 
		'LEU':['CB',1.0], 'LYS':['CB','CG','CD',3.0], 'GLU':['CB','CG',2.0], 'GLN':['CB','CG',2.0], 'ARG':['CB','CG','CD',3.0], 
		'HIS':['CB',1.0], 'PHE':['CB',1.0], 'TYR':['CB',1.0], 'TRP':['CB',1.0]}

    self.S2 = {'MET':['SD','CE',2.0], 'ASP':['CG','OD1','OD2',3.0], 'ASN':['CG','OD1','ND2',3.0], 'LEU':['CG','CD1','CD2',3.0], 'LYS':['CE','NZ',2.0],
		'GLU':['CD','OE1','OE2',3.0], 'GLN':['CD','OE1','NE2',3.0], 'ARG':['NE','CZ','NH1','NH2',4.0], 'HIS':['CG','ND1','CD2','CE1','NE2',5.0],
		'PHE':['CG','CD1','CD2','CE1','CE2','CZ',6.0], 'TYR':['CG','CD1','CD2','CE1','CE2','CZ','OH',7.0], 'TRP':['CG','CD1','CD2','NE1','CE2','CE3','CZ2','CZ3','CH2',9.0]}
    
    #BrutePDBParser.__init__(self)

  def un_atoms(self,diction,name,ua_name,atoms,res,atom_list,r):
    ''' This functions conwerts all atom representation to one united atom '''
    coord = array([0.0,0.0,0.0])
    n = 0.0 # do odejmowania od ilosc atomow, jesli jakiegos nie ma 
    for at in diction[name]:
	  if at in atoms:
	    ind = atoms.index(at)
	    coord += array(res.child_list[ind].get_coord(),'f')
	  elif type(at) == type(1.0): 
	    if (at-n) == 0.0: break
	    coord=coord/(at-n)
	    coor = [coord[0],coord[1],coord[2]]
	    try:
	    	a = Atom.Atom(ua_name,coor,0.0,1.0,' ',ua_name,1,ua_name[0])
	    except TypeError:
		a = Atom.Atom(ua_name,coor,0.0,1.0,' ',ua_name,1)
	    atom_list.append(a)
	    r.add(a)
	  else: 
	    print 'Missed atom %s in the %s %i' %(at,res.get_resname(), res.get_id()[1])
	    n +=1.0


  def get_prot_red(self,chains):
    atom_list = []
    res_num = 1
    for cha in chains:
	for res in cha.child_list:
	    atoms = []
	    name = res.get_resname()
	    r = Residue.Residue((' ',res_num, ' '), name, '')
	    res_num += 1
	    for at_name in res.child_list:
		atoms.append(at_name.get_name())
	    if 'CA' in atoms:
		ind = atoms.index('CA')
		coord = res.child_list[ind].get_coord()
		coor = [coord[0],coord[1],coord[2]]
		at_nam = res.child_list[ind].get_name()
		try:
		    a = Atom.Atom(at_nam,coor,0.0,1.0,' ',at_nam,1,at_nam[0])
	    	except TypeError:
		    a = Atom.Atom(at_nam,coor,0.0,1.0,' ',at_nam,1)
		atom_list.append(a)
		r.add(a)
	    else: print "Missed atom CA in the %s %i" %(name, res.get_id()[1])
	    if name in self.S1.keys():
		self.un_atoms(self.S1,name,'S1',atoms,res,atom_list,r)
	    elif name == 'GLY': None
	    if name in self.S2.keys():
		self.un_atoms(self.S2,name,'S2',atoms,res,atom_list,r)
    return atom_list


class RedRNA():
  '''convert RNA structure to reduced model according to MB. '''
  def __init__(self):
    self.same = { 'RIB': ["C4'","C3'","C1'","C2'","O4'", 5.0]}

    self.difer = {'R5':["N9","C8","N7","C5","C4",5.0], 'R6':["N1","C2","N3","C4","C6","C5",6.0], 'Y':["N1","C2","N3","C4","C6","C5", 6.0]}

    self.modyf = {'M5':["N9","C8","N7","C6","C5",5.0], 'M6':["N1","C2","N3","C4","C6","C5",6.0], 'MY':["N1","C2","N3","C4","C6","C5", 6.0]}
    
    #BrutePDBParser.__init__(self)

  def un_atoms(self,diction,ua_name,atoms,res,cha,atom_list,r):
    ''' This functions conwerts all atom representation to one united atom '''
    coord = array([0.0,0.0,0.0])
    n = 0.0 # do odejmowania od ilosc atomow, jesli jakiegos nie ma 
    for at in diction[ua_name]:
	  if at in atoms:
	    ind = atoms.index(at)
	    coord += array(res.child_list[ind].get_coord(),'f')
	  elif type(at) == type(1.0): 
	    if (at-n) == 0.0: break
	    coord=coord/(at-n)
	    coor = [coord[0],coord[1],coord[2]]
	    try:
	    	a = Atom.Atom(ua_name,coor,0.0,1.0,' ',ua_name,1,ua_name[0])
	    except TypeError:
		a = Atom.Atom(ua_name,coor,0.0,1.0,' ',ua_name,1)
	    atom_list.append(a)
	    r.add(a)
	  else: 
	    print 'Missed atom %s in the %s %i' %(at,res.get_resname(),res.get_id()[1])
	    n +=1.0


  def get_rna_red(self,chains):
    ''' convertion all atom RNA structure to reduced structure'''
    atom_list = []
    troj = {}
    res_num = 1
    for cha in chains:
	for res in cha.child_list:
	    atoms = []
	    name = res.get_resname()
	    r = Residue.Residue((' ',res_num, ' '), name, '')
	    naz = name.split()[0]+str(res_num)
	    res_num += 1
	    if "A" in name or "G" in name:
		try:
		    nazwa = {"H":res["N9"].get_coord(), "WC":res["C6"].get_coord(), "S":res["C2"].get_coord()}
		    troj[naz] = nazwa
		except KeyError: print "Warning, residue %s%i in the chain %s is not compleet" %(res.get_resname().split()[0], res.get_id()[1], cha.get_id())
	    elif "C" in name or "U" in name:
		try:
		    nazwa = {"H":res["C6"].get_coord(), "WC":res["C4"].get_coord(), "S":res["C2"].get_coord()}
		    troj[naz] = nazwa
		except KeyError: print "Warning, residue %s%i in the chain %s is not compleet" %(res.get_resname().split()[0], res.get_id()[1], cha.get_id())
	    else: print "Modyfikacja", naz
	    for at_name in res.child_list:
		atoms.append(at_name.get_name())
	    try:
		ind = atoms.index('P')
	    
		coord = res.child_list[ind].get_coord()
		coor = [coord[0],coord[1],coord[2]]
		at_nam = res.child_list[ind].get_name()
	    	try:
	    	    a = Atom.Atom(at_nam,coor,0.0,1.0,' ',at_nam,1,at_nam[0])
	    	except TypeError:
		    a = Atom.Atom(at_nam,coor,0.0,1.0,' ',at_nam,1)
		atom_list.append(a)
		r.add(a)
	    except ValueError: print "Warning, residue %s%i in the chain %s has not P atom" %(res.get_resname().split()[0], res.get_id()[1], cha.get_id())
	    
	    self.un_atoms(self.same,'RIB',atoms,res,cha,atom_list,r)
	    if "G" == name.strip() or name == "GUA" or "A" == name.strip() or name == "ADE":
		self.un_atoms(self.difer,'R5',atoms,res,cha,atom_list,r)
		self.un_atoms(self.difer,'R6',atoms,res,cha,atom_list,r)
	    elif "U" == name.strip() or name == "URI" or "C" == name.strip() or name == "CYT":
		self.un_atoms(self.difer,'Y',atoms,res,cha,atom_list,r)
	    elif "A" in name or "G" in name or name.strip() == "QUO":
		self.un_atoms(self.modyf,'M5',atoms,res,cha,atom_list,r)
		self.un_atoms(self.modyf,'M6',atoms,res,cha,atom_list,r)
	    elif "U" in name or "C" in name:
		self.un_atoms(self.modyf,'MY',atoms,res,cha,atom_list,r)
	    else: print "Unknown modification:", name, res.get_id()
    return atom_list, troj


f = open("energy_DARS_1.3.dat")
energy = pickle.load(f)
fang = open("energy_ang_DARS_1.3.dat")
energy_ang = pickle.load(fang)
fsite = open("energy_site_DARS_1.3.dat")
energy_site = pickle.load(fsite)
pars = BrutePDBParser()


def check_chains(plik, pars):
    """Check which chains are RNA and which are protin"""
    prot_names = ['ALA', 'VAL', 'LEU', 'ILE', 'MET', 'PHE', 'ASN', 'GLU', 'GLN', 'HIS', 'LYS', 'ARG', 'ASP', 'GLY', 'SER', 'THR', 'TYR', 'TRP', 'CYS', 'PRO']
    rna_names = ['A', 'C', 'G', 'U', 'ADE', 'CYT', 'GUA', 'URY', 'URI', 'U34', 'U31', 'C31', '4SU', 'H2U', 'QUO', 'G7M', '5MU', '5MC', 'PSU', '2MG', '1MG', '1MA', 'M2G', '5BU', 'FHU', 'FMU', 'IU', 'OMG', 'OMC', 'OMU', 'A2M', 'A23', 'CCC', 'I']
    count = 0
    prot_chains = []
    rna_chains = []
    struc = pars.get_structure('native', plik)
    chains = struc[0].child_list
    for ch in chains:
	prot_count = []
	rna_count = []
	for re in ch.child_list:
	    if re.get_resname().strip() != 'HOH':
		if re.get_resname().strip() in prot_names:
		    prot_count.append(1)
		elif re.get_resname().strip() in rna_names:
		    rna_count.append(1)
		else: 
		    print "Warning: ", re.get_resname().strip(), " is non-standard residue"
		    count += 1
	if sum(prot_count) > count and sum(rna_count) < sum(prot_count)/100.0:
	    prot_chains.append(ch)
	elif sum(rna_count) > count and sum(prot_count) < sum(rna_count)/100.0:
	    rna_chains.append(ch)
	else: print "Warning: %s has too much non-standard residues or it contains mixed protein and rna residues. Script does not know if it is protein or RNA. Tthis chain will not take into consideration during a model assessment." %ch
    return prot_chains, rna_chains


def reduced (infile, co):
    """ Reduce a ful atom representation """
    rna = RedRNA()
    prot = RedProt()
    print "file	E	Edis	Eang	Esite	Ep"		
    if co ==0:
	pliki = open(infile)
	pliki_tab = pliki.readlines()
	if len(pliki_tab)< opts.Nr_to_clust:
	    print "Error: number of structure to assess is less than number of clustering structures. Please use '-m' to define an appropriate number of clustering structures or '-n' to assess structures without clustering!!"
	    sys.exit(1)
	for plik in pliki_tab:
	    pro_ch, rna_ch = check_chains(plik.split()[0], pars)
	    rna_atoms, troj = rna.get_rna_red(rna_ch)
	    prot_atoms_list = prot.get_prot_red(pro_ch)
	    yield rna_atoms, prot_atoms_list, plik, troj    
    else:
	pro_ch, rna_ch = check_chains(infile.split()[0], pars)
	rna_atoms, troj = rna.get_rna_red(rna_ch)
	prot_atoms_list = prot.get_prot_red(pro_ch) 
	yield rna_atoms, prot_atoms_list, infile, troj
		
def ilwek (a,b):
  return (a[1]*b[2] - a[2]*b[1]) + (a[2]*b[0] - a[0]*b[2]) + (a[0]*b[1] - a[1]*b[0])

def calc_ang (one, two, three):
  a = one - two
  b = three - two
  an = Vector.norm(a)
  bn = Vector.norm(b)
  wekab = ilwek (a,b)
  sinab = wekab / (an*bn)
  cosab = (a*b)/(an*bn)
  ang = acos (cosab)
  if sinab<0: ang = ang+pi
  return degrees (ang)

def dis_site(lit):
    '''Calculate distances between protein united atom and all edges of nucleotide'''
    dif_lit = n.get_coord() - troj[rna_resid.get_resname().split()[0]+str(rna_resid.get_id()[1])][lit]
    pow_lit = dif_lit * dif_lit
    dis_lit = sqrt(sum(pow_lit))
    return dis_lit

def calculate_partial_RMSD(mat1, mat2, re, a):
    """
    Calculates RMSD for two residues.
    Returns RMSD.
    Arguments: list of fixed atoms, list of moved atoms, identifiers - string with names of residues
    """
    rmsd = 0.0
    if len(mat1) != len(mat2): print 'ERROR:Cannot calculate rmsd between two lists. Unequal lists lenght in the files %s and %s ' %(re, a)
    rmsd_mat = mat1 - mat2
    rmsd_mat = rmsd_mat**2
    rmsd = sqrt(rmsd_mat.sum()/len(rmsd_mat))
    return rmsd



Ener = {}
if opts.Structure:
    '''check if user give one structure or list of structures'''
    list_str = opts.Structure
    co = 1
else: 
    list_str = opts.List_of_structures
    co = 0


for rna_atoms, prot_atoms_list, plik, troj in reduced(list_str, co):
    Eang = 0.0
    Edis = 0.0
    Esite = 0.0
    Ekara = 0.0
    licz = 0.0
    #Potential calculating
    for rna_atom in rna_atoms:
	rna_one_atom = {}
	ns=NeighborSearch(prot_atoms_list)
	try: 
	    neighbors=ns.search(Numeric.array(rna_atom.get_coord(),"f"),9.0,'A')
	except TypeError:
	    neighbors=ns.search(array(rna_atom.get_coord()),9.0,'A')
	if neighbors != []:
	    for n in neighbors:
		licz = licz + 1.0
		residue_id=n.get_parent()
		rna_resid = rna_atom.get_parent()
		dis = sqrt(pow(rna_atom.get_coord()[0] - n.get_coord()[0], 2) + pow(rna_atom.get_coord()[1] - n.get_coord()[1], 2) + pow(rna_atom.get_coord()[2] - n.get_coord()[2], 2))
		if dis<3.0: 
		    if rna_atom.get_name() =="RIB" or rna_atom.get_name() =="P" or n.get_name()=="CA":
			Ekara = Ekara + 70000.0
			break
		elif dis<4.0:
		    if rna_atom.get_name() =="RIB" or rna_atom.get_name() =="P" or n.get_name()=="CA":
			Ekara = Ekara + 30000.0
			break
		wiez= "%s-%s:%s-%s_%i" %(rna_resid.get_resname().split()[0], rna_atom.get_name(), residue_id.get_resname(), n.get_name(), ceil(dis))
		if energy.has_key(wiez):
		    Edis = energy[wiez] + Edis
		if rna_atom.get_name()!="RIB":
		    if rna_resid.has_id("RIB"):		
			#print rna_resid, n, 'COS', rna_resid["RIB"].get_vector(), rna_atom.get_vector(),  n.get_vector()
			one = rna_resid["RIB"].get_vector()
			two = rna_atom.get_vector()
			three = n.get_vector()
			ang = calc_ang(one,two,three)
			#print ang
			if ang >=0.0 and ang <=20.0: angle = 20
			elif ang >20.0 and ang <=40.0: angle = 40
			elif ang >40.0 and ang <=60.0: angle = 60
			elif ang >60.0 and ang <=80.0: angle = 80
			elif ang >80.0 and ang <=100.0: angle = 100
			elif ang >100.0 and ang <=120.0: angle = 120
			elif ang >120.0 and ang <=140.0: angle = 140
			elif ang >140.0 and ang <=160.0: angle = 160
			elif ang >160.0 and ang <=180.0: angle = 180
			elif ang >180.0 and ang <=200.0: angle = 200
			elif ang >200.0 and ang <=220.0: angle = 220
			elif ang >220.0 and ang <=240.0: angle = 240
			elif ang >240.0 and ang <=260.0: angle = 260
			elif ang >260.0 and ang <=280.0: angle = 280
			elif ang >280.0 and ang <=300.0: angle = 300
			elif ang >300.0 and ang <=320.0: angle = 320
			elif ang >320.0 and ang <=340.0: angle = 340
			elif ang >340.0 and ang <=360.0: angle = 360
			wiez_ang = "%s-%s:%s-%s:%s-%s_%i" %(rna_resid.get_resname().split()[0], "RIB", rna_resid.get_resname().split()[0], rna_atom.get_name(), residue_id.get_resname(), n.get_name(), angle)
			if energy_ang.has_key(wiez_ang):
			    Eang = energy_ang[wiez_ang] + Eang
		elif "P" in rna_resid.child_dict:
		    one = rna_resid["P"].get_vector()
		    two = rna_atom.get_vector()
		    three = n.get_vector()
		    ang = calc_ang(one,two,three)
		    if ang >=0.0 and ang <=20.0: angle = 20
		    elif ang >20.0 and ang <=40.0: angle = 40
		    elif ang >40.0 and ang <=60.0: angle = 60
		    elif ang >60.0 and ang <=80.0: angle = 80
		    elif ang >80.0 and ang <=100.0: angle = 100
		    elif ang >100.0 and ang <=120.0: angle = 120
		    elif ang >120.0 and ang <=140.0: angle = 140
		    elif ang >140.0 and ang <=160.0: angle = 160
		    elif ang >160.0 and ang <=180.0: angle = 180
		    elif ang >180.0 and ang <=200.0: angle = 200
		    elif ang >200.0 and ang <=220.0: angle = 220
		    elif ang >220.0 and ang <=240.0: angle = 240
		    elif ang >240.0 and ang <=260.0: angle = 260
		    elif ang >260.0 and ang <=280.0: angle = 280
		    elif ang >280.0 and ang <=300.0: angle = 300
		    elif ang >300.0 and ang <=320.0: angle = 320
		    elif ang >320.0 and ang <=340.0: angle = 340
		    elif ang >340.0 and ang <=360.0: angle = 360
		    wiez_ang = "%s-%s:%s-%s:%s-%s_%i" %(rna_resid.get_resname().split()[0], "P", rna_resid.get_resname().split()[0], rna_atom.get_name(), residue_id.get_resname(), n.get_name(), angle)
		    if energy_ang.has_key(wiez_ang):
			Eang = energy_ang[wiez_ang] + Eang

		if rna_atom.get_name() =="R5" or rna_atom.get_name() =="R6" or rna_atom.get_name() =="Y":
		    try:
			dis_h = dis_site("H")
			dis_wc = dis_site("WC")
			dis_s = dis_site("S")
			if dis_wc < dis_s < dis_h or dis_s < dis_wc < dis_h:
			    wiez_site = "%s:%s-%s_WoCr" % (rna_resid.get_resname().split()[0],residue_id.get_resname(), n.get_name())
			    if energy_site.has_key(wiez_site):
				Esite = energy_site[wiez_site] + Esite
			elif dis_h < dis_s < dis_wc or dis_s < dis_h < dis_wc:
			    wiez_site = "%s:%s-%s_Sug" % (rna_resid.get_resname().split()[0],residue_id.get_resname(), n.get_name())
			    if energy_site.has_key(wiez_site):
				Esite = energy_site[wiez_site] + Esite
			elif dis_h < dis_wc < dis_s or dis_wc < dis_h < dis_s:
			    wiez_site = "%s:%s-%s_Hoo" % (rna_resid.get_resname().split()[0],residue_id.get_resname(), n.get_name())
			    if energy_site.has_key(wiez_site):
				Esite = energy_site[wiez_site] + Esite
			else: print "Problem with edges of nucleotide!!"
		    except KeyError: print "Residue %s %i is incomplete - edges of nucleotide can not be calculated." %(rna_resid.get_resname().split()[0], rna_resid.get_id()[1])

    E = Edis + Eang + Esite + Ekara
    print "%s	%.2f	%.2f	%.2f	%.2f	%.2f" %(plik.split("/")[-1].split()[0], E, Edis, Eang, Esite, Ekara)
    #print plik.split("/")[-1].split()[0], "	", E
    Ener[E] = plik.split("/")[-1].split()[0]

if opts.Structure or opts.No_clustering:
    sys.exit(1)

lent = len(plik.split("/")[-1].split()[0])
pref = plik.split()[0][:-(lent)]

# Sorting of the best scored structures
nr_clu = opts.Nr_to_clust
print "%i best scored structures" % nr_clu
best_naz = []
score = []
best_key = Ener.keys()
best_key.sort()
the_best = best_key[0:nr_clu]
for best in the_best:
    print Ener[best], "	", best
    best_naz.append(Ener[best])
    score.append(best)

#Clustering procedure, proposed by Baker
size = len(the_best)
matr = zeros((size,size))
matri = zeros((size,size))

cf = opts.Cut_off
for re in best_naz:
    atom_ref = []
    plik_re = pref+re
    ref = open(plik_re)
    for line in ref.readlines():
	if "ATOM" in line:
	    if line[21:22] == "B":
		coord = []
		coord.append(float(line[30:38]))
		coord.append(float(line[38:46]))
		coord.append(float(line[46:54]))
		atom_ref.append(coord)
    ref.close()
    mat1 = array(atom_ref)
    for a in best_naz:
	plik_a = pref+a
	chan = open(plik_a)
	atom_change = []
	for line in chan.readlines():
	    if "ATOM" in line:
		if line[21] == "B":
		    coord = []
		    coord.append(float(line[30:38]))
		    coord.append(float(line[38:46]))
		    coord.append(float(line[46:54]))
		    atom_change.append(coord)
	mat2 = array(atom_change)
	rms   = calculate_partial_RMSD(mat1, mat2, re, a)
	matri[best_naz.index(re), best_naz.index(a)] = rms
	if rms < cf: matr[best_naz.index(re.split()[0]), best_naz.index(a.split()[0])] = 1
	else: matr[best_naz.index(re.split()[0]), best_naz.index(a.split()[0])] = 0
	chan.close()

# Print only clusters with more then 4 structures in it
print "CLUSTERS, %i best scored models, cut off %i A" %(opts.Nr_to_clust, opts.Cut_off)
while 1:
    biggest = []

    for a in matr:
	licz = a.sum()
	biggest.append(licz)

    big_est = array(biggest)
    maxim = big_est.max()
    if maxim < 5: break
    ind = biggest.index(maxim)
    niez = matr[ind,:]
    tozer = niez.nonzero()
    
    print len(tozer[0])
    clust = []
    for ze in tozer[0]:
	clust.append(best_naz[ze])
	nam = best_naz[ze]
	print nam, "	", score[best_naz.index(nam)]#, "	", matri[best_naz.index(nam), best_naz.index('gramm_10000/natywna.pdb')]

    matr[:,tozer]= 0


	  
	  
